from os import PathLike
from typing_extensions import Self
from abc import abstractmethod, ABC as _ABC
from numpy import sin as _sin, cos as _cos
from typing import Any, Union, Sequence, Iterable

from papercandy.universal import utils as _utils


class DataCompound(object):
    def __init__(self, data: Any, target: Any, d_type: type = Any):
        self.data: d_type = data
        self.target: d_type = target

    @abstractmethod
    def cuda(self) -> Self:
        raise NotImplementedError

    def unpack(self) -> [Any, Any]:
        return self.data, self.target


class Container(object):
    @abstractmethod
    def __init__(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def cuda(self) -> Self:
        raise NotImplementedError

    @abstractmethod
    def get(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def save(self, path: Union[str, PathLike]):
        raise NotImplementedError


class LayerInfo(object):
    def __init__(self, width: int, height: int, angle: int, name: str, description: str = ""):
        self.width: int = width
        self.g_width: int = round(_cos(_utils.angle2radian(angle)) * width)
        self.height: int = height
        self.angle: int = angle
        self.name: str = name
        self.description: str = description

    def get(self) -> (int, int, int, int, str, str):
        return self.width, self.g_width, self.height, self.angle, self.name, self.description

    def parse_interval(self, interval: Union[int, float]) -> int:
        if interval >= 1:
            return interval
        return interval * self.g_width


class LayerInfoList(Sequence):
    def __init__(self, *layers: LayerInfo):
        self._layers: list[LayerInfo] = list(layers)

    def __iter__(self) -> Iterable[LayerInfo]:
        return iter(self._layers)

    def __len__(self) -> int:
        return len(self._layers)

    def __getitem__(self, item: Union[int, slice]) -> Union[LayerInfo, list[LayerInfo]]:
        return self._layers[item]

    def __call__(self, interval: Union[int, float]) -> (int, int):
        if interval < 0:
            raise ValueError("`interval` cannot be negative.")
        if len(self._layers) < 1:
            raise IndexError("There must be at least one layer.")
        canvas_width = 0
        canvas_height = 0
        display_interval = None
        for layer in self._layers:
            display_interval = layer.parse_interval(interval)
            canvas_width += layer.g_width + display_interval
            graph_height = layer.width * _sin(_utils.angle2radian(layer.angle)) + layer.height
            if graph_height > canvas_height:
                canvas_height = graph_height
        canvas_width -= display_interval
        return round(canvas_width), round(canvas_height)

    def __add__(self, other) -> Self:
        if not isinstance(other, LayerInfoList):
            raise TypeError("LayerInfoList can only be added to another LayerInfoList.")
        return LayerInfoList(*(self._layers + other.get_layers()))

    def reverse(self) -> Self:
        self._layers.reverse()
        return self

    def get_layers(self) -> list[LayerInfo]:
        return self._layers

    def append(self, layer_info: LayerInfo) -> Self:
        self._layers.append(layer_info)
        return self


class NetworkC(Container, _ABC):
    def __len__(self) -> int:
        return len(self.structure())

    @abstractmethod
    def structure(self) -> LayerInfoList[LayerInfo]:
        raise NotImplementedError


class LossFunctionC(Container, _ABC):
    pass


class OptimizerC(Container, _ABC):
    pass



from typing import Any
from torch import Tensor as _Tensor
from torch.nn import Module as _Module

from papercandy import network as _network
from papercandy.universal import train as _train


class TrainingMonitor(_train.TrainingMonitor):
    def on_updated(self, epoch: int, loss: float, input_data: _network.DataCompound, output: _Tensor): pass


class Trainer(_train.Trainer):
    def _train_one_batch(self, epoch: int, network: _Module, loss_function: _Module, optimizer: _Module,
                         data: _network.DataCompound) -> [Any, float]:
        optimizer.zero_grad()
        output = network(data.data)
        current_loss = loss_function(output, data.target)
        current_loss.backward()
        optimizer.step()
        return output, current_loss.item()



from os import PathLike
from typing import Union
from torch.cuda import is_available as _is_available

from papercandy.universal import config as _config

Config = _config.Config
Bool = _config.Bool


def new_config(filename: Union[str, PathLike]) -> Config:
    cfg = _config.new_config(filename)
    if not _is_available():
        cfg.put("gpu_acceleration", "False")
    return cfg



from typing import Union, Any
from abc import abstractmethod

from papercandy import network as _network
from papercandy.universal import dataloader as _dl, config as _cfg


class TrainingMonitor(object):
    def on_updated(self, epoch: int, loss: float, input_data: _network.DataCompound, output: Any): pass


class Trainer(object):
    def __init__(self, config: _cfg.Config = None, dataloader: _dl.Dataloader = None):
        self._nc: Union[_network.NetworkC, None] = None
        self._lfc: Union[_network.LossFunctionC, None] = None
        self._oc: Union[_network.OptimizerC, None] = None
        self._config: Union[_cfg.Config, None] = config
        self._dataloader: Union[_dl.Dataloader, None] = dataloader
        self._epoch: int = 0
        self.losses: list[float] = []

    def _check_requirements(self) -> bool:
        return self._config is not None and self._dataloader is not None and self._nc is not None \
               and self._lfc is not None and self._oc is not None

    def _check_requirements_and_raise_exception(self):
        if not self._check_requirements():
            raise AttributeError("Trainer hasn't been completely prepared.")

    def get_config(self) -> _cfg.Config:
        return self._config

    def get_dataloader(self) -> _dl.Dataloader:
        return self._dataloader

    def get_epoch(self) -> int:
        return self._epoch

    def set_network(self, nc: _network.NetworkC):
        if self._config.get_predefined("cuda_acceleration", True):
            nc = nc.cuda()
        self._nc = nc

    def get_network(self) -> Union[_network.NetworkC, None]:
        return self._nc

    def set_loss_function(self, lfc: _network.LossFunctionC):
        if self._config.get_predefined("cuda_acceleration", True):
            lfc = lfc.cuda()
        self._lfc = lfc

    def get_loss_function(self) -> Union[_network.LossFunctionC, None]:
        return self._lfc

    def set_optimizer(self, oc: _network.OptimizerC):
        if self._config.get_predefined("cuda_acceleration", True):
            oc = oc.cuda()
        self._oc = oc

    def get_optimizer(self) -> Union[_network.OptimizerC, None]:
        return self._oc

    def train(self, num_batches: int = 1, monitor: TrainingMonitor = TrainingMonitor()):
        self._check_requirements_and_raise_exception()
        epoch = 0
        for data in self.get_dataloader():
            if epoch >= num_batches:
                break
            if self._config.get_predefined("cuda_acceleration", True):
                data = data.cuda()
            self.train_one_batch(data, monitor)
            epoch += 1
        self._epoch += epoch

    def train_one_batch(self, data_batch: _network.DataCompound, monitor: TrainingMonitor):
        self._check_requirements_and_raise_exception()
        o, loss = self._train_one_batch(self.get_epoch(), self.get_network().get(), self.get_loss_function().get(),
                                        self.get_optimizer().get(), data_batch)
        self.losses.append(loss)
        monitor.on_updated(self.get_epoch(), loss, data_batch, o)

    @abstractmethod
    def _train_one_batch(self, epoch: int, network: Any, loss_function: Any, optimizer: Any,
                         data: _network.DataCompound) -> [Any, float]:
        """
        :param epoch: global epoch
        :param network: network (not container)
        :param loss_function: loss function (not container)
        :param optimizer: optimizer (not container)
        :param data: single data, not batched
        :return: network output, loss
        """
        raise NotImplementedError



from papercandy.train import *
from papercandy.config import *
from papercandy.drawing import *
from papercandy.network import *
from papercandy.version import *
from papercandy.dataloader import *


CONFIG = None


def load_config(filename: Union[str, PathLike]):
    global CONFIG
    CONFIG = new_config(filename)



from typing import Union
from typing_extensions import Self
from torch import Tensor as _Tensor
from os import listdir as _listdir, PathLike


from papercandy import network as _network
from papercandy.universal import dataloader as _dataloader


Dataset = _dataloader.Dataset


class Dataloader(_dataloader.Dataloader):
    def combine_batch(self, data_batch: list[_network.DataCompound]) -> _network.DataCompound:
        d_list, t_list = [], []
        for dc in data_batch:
            d, t = dc.unpack()
            d, t = d.tolist(), t.tolist()
            d_list.append(d)
            t_list.append(t)
        return _network.DataCompound(_Tensor(d_list), _Tensor(t_list))


class ExampleDataset(Dataset):
    def __init__(self, src: Union[str, PathLike]):
        self.src: Union[str, PathLike] = src
        self.file_list: list = _listdir(self.src)

    def __len__(self) -> int:
        return len(self.file_list)

    def cut(self, i: slice) -> Self:
        o = ExampleDataset(self.src)
        o.file_list = self.file_list[i]
        return o

    def get(self, i: int) -> _network.DataCompound:
        with open("%s/%s" % (self.src, self.file_list[i]), "r") as f:
            return _network.DataCompound(_Tensor([int(self.file_list[i])]), _Tensor(eval(f.read())))



import setuptools

setuptools.setup(
    name="paper-candy",
    version="1.0.0b1",
    author="ATATC",
    author_email="futerry@outlook.com",
    description="A loosely coupled lightweight framework for deep learning papers.",
    license='Apache License 2.0',
    long_description="**PaperCandy** is a loosely coupled lightweight framework for deep learning papers. "
                     "It provides a series of auxiliary tools for rapid rebuilding and writing papers, "
                     "including universal data loader, and automatically generating network structure and "
                     "training process diagram. So far support PyTorch as the only front-end framework.",
    long_description_content_type="text/markdown",
    url="https://github.com/ATATC/PaperCandy",
    packages=setuptools.find_packages(),
    install_requires=["torch", "opencv-python", "matplotlib"],
)
